# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (C) 2012 Attila Siposs <sipiatti@gmail.com>
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

import gettext
import sqlite3
import os, sys

import logging
logger = logging.getLogger('pynote_lib')

from gettext import gettext as _
gettext.textdomain('pynote')

from gi.repository import Gtk
from gi.repository import Gio

class PynoteTreeStorage(Gtk.TreeStore):
    """Base pynote storage system. It holds the categories and notes hierarchy."""

    def __init__(self,*args):
        super(PynoteTreeStorage,self).__init__(int,int,str,int) # id, type, title/name, parent

        # checking if storage file is set in preferences
        self.settings = Gio.Settings("net.launchpad.pynote")
        self.settings.connect("changed::storagefile", self.on_storagefile_changed)
        storagefile = self.settings.get_string("storagefile")
        self.createcategorysql = u"create table category (id integer primary key autoincrement, name text, parent int)"
        self.createnotessql = u"create table notes (id integer primary key autoincrement, title text, category_id int, content text)"
        self.createcategorysql_ne = u"create table if not exists category (id integer primary key autoincrement, name text, parent int)"
        self.createnotessql_ne = u"create table if not exists notes (id integer primary key autoincrement, title text, category_id int, content text)"
        if storagefile == '':
            storagefile = os.path.join(os.path.expanduser('~'),'.pynote/notes.db')
            self.settings.set_string("storagefile",storagefile)
        else:
            self.on_storagefile_changed(storagefile)

    def on_storagefile_changed(self,a=None,b=None):
        storagefile = self.settings.get_string("storagefile")

        # if directory does not exist create it
        if not os.path.exists(os.path.dirname(storagefile)):
            # need try-except block
            os.makedirs(os.path.dirname(storagefile))

        # if storagefile does not exist create new storagedb, else check if its format is OK
        if not os.path.exists(storagefile):
            logger.debug('Storagefile does not exists: %s' % storagefile)
            self.createnotedb(storagefile)
        else:
            self.checknotedb(storagefile)

        self.fill_tree_from_db()

    def createnotedb(self,storagefile):
        self.db = sqlite3.connect(storagefile)
        self.dbsql = self.db.cursor()
        self.dbsql.execute(self.createcategorysql_ne)
        self.dbsql.execute(self.createnotessql_ne)
        logger.debug('New db file created: %s' % storagefile)

    def checknotedb(self,storagefile):
        try:
            self.db = sqlite3.connect(storagefile)
            self.dbsql = self.db.cursor()
            self.dbsql.execute("select * from sqlite_master where name='category' or name='notes'")
        except sqlite3.DatabaseError:
            logger.debug('The storagefile (%s) is not an SQLite db file. Reverting back to default' % storagefile)
            dialog = Gtk.MessageDialog(None, 0, Gtk.MessageType.ERROR,Gtk.ButtonsType.OK, _("Storagefile %s is invalid") % storagefile)
            dialog.format_secondary_text(_("The given file is not a valid pynote datafile.\nReverting back to the default."))
            dialog.run()
            dialog.destroy()

            storagefile = os.path.join(os.path.expanduser('~'),'.pynote/notes.db')
            self.createnotedb(storagefile)
            self.settings.set_string("storagefile",storagefile)
            return

        else:
            result = self.dbsql.fetchall()
            if len(result) == 2 and result[0][4].lower() == self.createcategorysql and result[1][4].lower() == self.createnotessql:
                logger.debug('Storagefile %s is valid' % storagefile)
            else:
                logger.debug('Storagefile %s is invalid' % storagefile)
                if storagefile == os.path.join(os.path.expanduser('~'),'.pynote/notes.db'):
                    dialog = Gtk.MessageDialog(None, 0, Gtk.MessageType.ERROR,Gtk.ButtonsType.OK, _("Storagefile %s is invalid") % storagefile)
                    dialog.format_secondary_text(_("The default pynote datafile is corrupted and will be deleted.\nNew datafile will be created."))
                    dialog.run()
                    dialog.destroy()
                    os.remove(storagefile)
                else:
                    dialog = Gtk.MessageDialog(None, 0, Gtk.MessageType.ERROR,Gtk.ButtonsType.OK, _("Storagefile %s is invalid") % storagefile)
                    dialog.format_secondary_text(_("The given file is not a valid pynote datafile.\nReverting back to the default."))
                    dialog.run()
                    dialog.destroy()

                storagefile = os.path.join(os.path.expanduser('~'),'.pynote/notes.db')
                self.createnotedb(storagefile)
                self.settings.set_string("storagefile",storagefile)

    def fill_tree_from_db(self):
        self.clear()
        # default uncategorized category
        self.uncategorized = self.append(None,[0,1,_("Uncategorized"),0])
        self.catpointers = {0:self.uncategorized}
        self.notepointers = {}
        self.note_contents = {}
        self.dbsql.execute("select * from category order by name")
        result = self.dbsql.fetchall()
        for row in result:
            self.catpointers[row[0]] = self.append(None,[row[0],1,row[1],0])
        self.dbsql.execute("select * from notes order by title")
        result = self.dbsql.fetchall()
        for row in result:
            self.note_contents[row[0]] = row[3]
            parent = self.catpointers[0]
            if row[2] != None and row[2] != '':
                parent = self.catpointers[row[2]]
            self.notepointers[row[0]] = self.append(parent,[row[0],0,row[1],row[2]])
        logger.debug('Note tree filled up')

    def add_new_category(self,name):
        self.dbsql.execute("insert into category (name) values (?)",(unicode(name.decode("utf-8")),))
        self.db.commit()
        rowid = self.dbsql.lastrowid
        self.dbsql.execute("select * from category order by name")
        result = self.dbsql.fetchall()
        new_iter = self.append(None,[rowid,1,name,0])
        self.catpointers[rowid] = new_iter
        cur_iter = self.iter_next(self.get_iter_first())
        while cur_iter != None:
            if cur_iter != new_iter:
                if name < self[cur_iter][2]:
                    self.move_before(new_iter,cur_iter)
                    break
            cur_iter = self.iter_next(cur_iter)

    def modify_category(self,iter,name):
        rowid = self[iter][0]
        self.dbsql.execute("update category set name=? where id=?",(unicode(name.decode("utf-8")),rowid))
        self.db.commit()
        self[iter][2] = name
        self.move_before(iter,None)
        cur_iter = self.iter_next(self.get_iter_first())
        while cur_iter != None:
            if cur_iter != iter:
                if name < self[cur_iter][2]:
                    self.move_before(iter,cur_iter)
                    break
            cur_iter = self.iter_next(cur_iter)

    def delete_category(self,iter):
        category_id = self[iter][0]
        self.dbsql.execute("delete from category where id=?",(category_id,))
        self.db.commit()
        self.remove(iter)
        del(self.catpointers[category_id])

    def delete_note(self,iter):
        note_id = self[iter][0]
        self.dbsql.execute("delete from notes where id=?",(note_id,))
        self.db.commit()
        self.remove(iter)
        del(self.notepointers[note_id])

    def count_children(self,category_id):
        self.dbsql.execute("select count(id) from notes where category_id=?",(category_id,))
        result = self.dbsql.fetchone()
        return result[0]

    def add_new_note(self, name, category_id):
        self.dbsql.execute("insert into notes (title,category_id,content) values (?,?,?)",(unicode(name.decode("utf-8")),category_id,""))
        self.db.commit()
        rowid = self.dbsql.lastrowid
        parent = self.catpointers[category_id]
        new_iter = self.append(parent,[rowid,0,name,category_id])
        self.notepointers[rowid] = new_iter
        self.note_contents[rowid] = ""
        cur_iter = self.iter_children(parent)
        while cur_iter != None:
            if cur_iter != new_iter:
                if name < self[cur_iter][2]:
                    self.move_before(new_iter,cur_iter)
                    break
            cur_iter = self.iter_next(cur_iter)
        return new_iter

    def save_current_note(self,note_id,note_title,note_content,category_id):
        self.dbsql.execute("update notes set title=?, content=?, category_id=? where id=?",(unicode(note_title.decode("utf-8")), unicode(note_content.decode("utf-8")), category_id, note_id))
        self.db.commit()
        self.note_contents[note_id] = note_content
        need_reorder = False
        current_note_iter = self.notepointers[note_id]
        current_note_title = self[current_note_iter][2]
        if current_note_title != note_title:
            self[current_note_iter][2] = note_title
            need_reorder = True
        current_note_category = self[current_note_iter][3]
        if current_note_category != category_id:
            new_iter = self.append(self.catpointers[category_id], [note_id,0,note_title,category_id])
            self.notepointers[note_id] = new_iter
            self.remove(current_note_iter)
            current_note_iter = new_iter
            need_reorder = True
        if need_reorder:
            cur_iter = self.iter_children(self.catpointers[category_id])
            self.move_before(current_note_iter,None)
            while cur_iter != None:
                if cur_iter != current_note_iter:
                    if note_title < self[cur_iter][2]:
                        self.move_before(current_note_iter,cur_iter)
                        break
                cur_iter = self.iter_next(cur_iter)
